from __future__ import division

import copy
import itertools
import math
import json
import logging

__author__ = 'vahram'

#==================================================================================================#
# Logging

logger = logging.getLogger('bayesnet')

#==================================================================================================#
# Exception Class

class BayesNetException(Exception):
    pass

#==================================================================================================#
# Model Input Class

class VariableInfo(object):

    def __init__(self, var_names):

        self.var_names = var_names
        self.var_defs = {v: {"type": "str",
                             "vals": None,
                             "priors": None,
                             "parents": [],
                             "children": []}
                         for v in var_names}
        self.is_ordered = False

        logger.info("Variable names loaded: %s", var_names)

    def set_variable_values(self, var_name, vals):

        var_type = "str"
        if vals:
            if isinstance(vals[0], bool):
                var_type = "bool"
            elif isinstance(vals[0], int):
                var_type = "int"
            elif isinstance(vals[0], float):
                var_type = "float"

        self.var_defs[var_name]["type"] = var_type
        self.var_defs[var_name]["vals"] = vals
        self.var_defs[var_name]["priors"] = {val: 1.0 for val in vals}

        logger.info("Variable values loaded for %s: %s", var_name, vals)

    def set_variable_priors(self, var_name, prior_counts):

        for val in self.var_defs[var_name]["vals"]:
            self.set_variable_single_prior(var_name, val, prior_counts[val])

    def set_variable_single_prior(self, var_name, val, prior_count):

        if prior_count < 0.0:
            raise BayesNetException("Count cannot be negative")

        self.var_defs[var_name]["priors"][val] = prior_count

        logger.info("Variable prior count loaded for %s=%s: %s", var_name, val, prior_count)

    def add_relationship(self, var_parent, var_child):

        self.var_defs[var_parent]["children"].append(var_child)
        self.var_defs[var_child]["parents"].append(var_parent)

        logger.info("Variable relationship added: %s -> %s", var_parent, var_child)

    def finalize(self):

        self.var_names = self._topological_sort()
        self.is_ordered = True

        logger.info("Variable info finalized: %s", self.var_names)

    def validate(self, exception_on_error=True):

        vals_ok = all([isinstance(v_def["vals"], list)
                      and len(v_def["vals"]) > 0
                      for (v_name, v_def) in self.var_defs.iteritems()])
        if not vals_ok and exception_on_error:
            raise BayesNetException("Validation failed on variable values")

        priors_ok = all([isinstance(v_def["priors"], dict) and
                        set(v_def["priors"].keys()) == set(v_def["vals"]) and
                        all([c >= 0.0 for c in v_def["priors"].values()])
                        for (v_name, v_def) in self.var_defs.iteritems()])
        if not priors_ok and exception_on_error:
            raise BayesNetException("Validation failed on variable priors")

        parents_ok = all([isinstance(v_def["parents"], list) and
                         all([v in self.var_defs for v in v_def["parents"]])
                         for (v_name, v_def) in self.var_defs.iteritems()])
        if not parents_ok and exception_on_error:
            raise BayesNetException("Validation failed on variable parents")

        children_ok = all([isinstance(v_def["children"], list) and
                          all([v in self.var_defs for v in v_def["children"]])
                          for (v_name, v_def) in self.var_defs.iteritems()])
        if not children_ok and exception_on_error:
            raise BayesNetException("Validation failed on variable children")

        if not self.is_ordered and exception_on_error:
            raise BayesNetException("Validation failed on variable ordering")

        return vals_ok and priors_ok and parents_ok and children_ok and self.is_ordered

    #----------------------------------------------------------------------------------------------#
    # Private

    def _topological_sort(self):

        graph = {v: set(v_def["parents"]) for (v, v_def) in self.var_defs.iteritems()}
        final = []

        while True:
            roots = set(v for (v, parents) in graph.iteritems() if not parents)
            if not roots:
                break

            final += list(roots)
            graph = {v: (parents - roots)
                     for (v, parents) in graph.iteritems()
                     if v not in roots}

        if any([len(parents) > 0 for parents in graph.itervalues()]):
            raise BayesNetException("Detected a cycle in the graph")

        assert set(final) == set(self.var_names), "Topological sort is broken"
        return final


#==================================================================================================#
# Helper Factor Class

class Factor(object):

    def __init__(self, scope_vars, vals_to_probs):

        self.scope_vars = scope_vars
        self.vals_to_probs = vals_to_probs

        logger.info("New factor initialized: scope=%s", self.scope_vars)

    def __repr__(self):

        return "Factor(%s, %s)" % (self.scope_vars, self.vals_to_probs)

    def copy(self):

        return Factor(copy.copy(self.scope_vars), copy.copy(self.vals_to_probs))

    def sum_out(self, var_out, inplace=True):

        i_out = self.scope_vars.index(var_out)

        scope_vars_new = self.scope_vars[:i_out] + self.scope_vars[i_out + 1:]
        vals_to_probs_new = {}

        for vals, prob in self.vals_to_probs.iteritems():
            vals_new = vals[:i_out] + vals[i_out + 1:]
            vals_to_probs_new[vals_new] = vals_to_probs_new.get(vals_new, 0.0)
            vals_to_probs_new[vals_new] += prob

        logger.info("Factor sum out of %s complete: scope=%s", var_out, scope_vars_new)

        return self._process_update(scope_vars_new, vals_to_probs_new, inplace=inplace)

    def max_out(self, var_out, inplace=True):

        i_out = self.scope_vars.index(var_out)

        scope_vars_new = self.scope_vars[:i_out] + self.scope_vars[i_out + 1:]
        vals_to_probs_new = {}

        for vals, prob in self.vals_to_probs.iteritems():
            vals_new = vals[:i_out] + vals[i_out + 1:]
            if vals_new not in vals_to_probs_new or vals_to_probs_new[vals_new] < prob:
                vals_to_probs_new[vals_new] = prob

        logger.info("Factor max out of %s complete: scope=%s", var_out, scope_vars_new)

        return self._process_update(scope_vars_new, vals_to_probs_new, inplace=inplace)

    def arg_max(self, var_arg, rest_var_to_val):

        i_arg = self.scope_vars.index(var_arg)
        rest_vals = [rest_var_to_val[var] for var in self.scope_vars if var != var_arg]

        max_val = None
        max_prob = -1

        for vals, prob in self.vals_to_probs.iteritems():
            arg_cur = vals[i_arg]
            vals_cur = vals[:i_arg] + vals[i_arg + 1:]
            assert len(vals_cur) == len(rest_vals), "Max-product traceback is broken"

            if vals_cur == tuple(rest_vals):
                logger.debug("\tProcessing arg=%s, rest=%s", arg_cur, vals_cur)

                if prob > max_prob:
                    max_prob = prob
                    max_val = arg_cur

        logger.info("Arg max of %s complete (scope=%s): %s -> %.4f",
                    var_arg, self.scope_vars, max_val, max_prob)
        return max_val

    def product(self, f, inplace=True):

        i_new = [i for (i, v) in enumerate(f.scope_vars) if v not in self.scope_vars]
        i_j_common = [(self.scope_vars.index(v), i)
                      for (i, v) in enumerate(f.scope_vars) if v in self.scope_vars]

        scope_vars_new = self.scope_vars + [f.scope_vars[i] for i in i_new]
        vals_to_probs_new = {}

        logger.debug("Running factor product on:\n%s\n%s", self, f)
        logger.debug("New scope variables: %s", scope_vars_new)
        logger.debug("Common index pairs: %s", i_j_common)
        logger.debug("Second factor new indices: %s", i_new)

        for vals, prob in self.vals_to_probs.iteritems():
            for f_vals, f_prob in f.vals_to_probs.iteritems():
                if all([vals[i] == f_vals[j] for i, j in i_j_common]):
                    vals_new = vals + tuple([f_vals[i] for i in i_new])
                    prob_new = prob * f_prob
                    vals_to_probs_new[vals_new] = prob_new

        logger.info("Factor product complete: scope=%s", scope_vars_new)

        return self._process_update(scope_vars_new, vals_to_probs_new, inplace=inplace)

    def reduce(self, var, val, inplace=True):

        try:
            i_var = self.scope_vars.index(var)
        except ValueError:
            if inplace:
                return self
            else:
                return self.copy()

        scope_vars_new = self.scope_vars[:i_var] + self.scope_vars[i_var + 1:]
        vals_to_probs_new = {}

        for vals, prob in self.vals_to_probs.iteritems():
            if vals[i_var] == val:
                vals_new = vals[:i_var] + vals[i_var + 1:]
                vals_to_probs_new[vals_new] = prob

        logger.info("Factor reduction by %s=%s complete: scope=%s", var, val, scope_vars_new)

        return self._process_update(scope_vars_new, vals_to_probs_new, inplace=inplace)

    def normalize(self, inplace=True):

        sum_prob = sum(self.vals_to_probs.itervalues())
        assert sum_prob > 0.0, \
               "Inference is broken: scope=%s, sum=%f" % (self.scope_vars, sum_prob)

        scope_vars_new = copy.copy(self.scope_vars)
        vals_to_probs_new = {vals: (prob / sum_prob)
                             for (vals, prob) in self.vals_to_probs.iteritems()}

        logger.info("Factor normalization complete: scope=%s", scope_vars_new)

        return self._process_update(scope_vars_new, vals_to_probs_new, inplace=inplace)

    #----------------------------------------------------------------------------------------------#
    # Private

    def _process_update(self, scope_vars_new, vals_to_probs_new, inplace=True):

        # This can get huge, uncomment with caution:
        logger.debug("Factor update: scope=%s\n%s", scope_vars_new, vals_to_probs_new)

        if inplace:
            self.scope_vars = scope_vars_new
            self.vals_to_probs = vals_to_probs_new
            return self

        return Factor(scope_vars_new, vals_to_probs_new)

#==================================================================================================#
# Main Model Class

class BayesNet(object):

    #----------------------------------------------------------------------------------------------#
    # Helpers

    @staticmethod
    def pandas_dataframe_to_table(df):
        return [row.to_dict() for _, row in df.iterrows()]

    @staticmethod
    def get_values_from_table(table):
        return {col: sorted(set(row[col] for row in table))
                for col in table[0].keys()}

    @staticmethod
    def clean_table(table, var_info):
        return [rec for rec in table
                if all([rec[x] in var_info.var_defs[x]["vals"]
                        for x in rec.keys()
                        if x in var_info.var_defs])]

    #----------------------------------------------------------------------------------------------#
    # Initialization

    def __init__(self, **kwargs):

        self.var_names = None
        self.var_defs = None
        self.probs = None

        self.factors = None
        self.cached_factors = None

        self.log_mode = kwargs.get("log_mode", False)

        logger.info("Initialized empty bayesnet")

    #----------------------------------------------------------------------------------------------#
    # Training

    def train(self, var_info, table, row_counts=None):

        logger.info("Training bayesnet: (%d vars, %d rows)", len(var_info.var_names), len(table))

        var_info.validate(exception_on_error=True)

        self.var_names = var_info.var_names
        self.var_defs = var_info.var_defs

        self.factors = None
        self.cached_factors = None

        self.probs = {v: {"counts": {}, "totals": {}, "probs": {}}
                      for v in self.var_names}

        for v in self.probs.iterkeys():
            self._init_counts(v, self.probs[v]["counts"], self.probs[v]["totals"], "", 0)

        for i, row in enumerate(table):
            for v in self.var_names:
                row_val = row[v]
                key = "|".join([str(row[v_parent])
                               for v_parent in self.var_defs[v]["parents"]])

                assert key in self.probs[v]["counts"], \
                       "Train is broken: key=%s, counts=%s" % (key, self.probs[v]["counts"])
                assert key in self.probs[v]["totals"], \
                       "Train is broken: key=%s, totals=%s" % (key, self.probs[v]["totals"])

                if not row_counts:
                    self.probs[v]["counts"][key][row_val] += 1
                    self.probs[v]["totals"][key] += 1
                else:
                    self.probs[v]["counts"][key][row_val] += row_counts[i]
                    self.probs[v]["totals"][key] += row_counts[i]

        logger.info("Counts and totals updated from data")

        for v in self.var_names:
            priors = self.var_defs[v]["priors"]
            sum_priors = sum(prior for prior in priors.itervalues())

            assert sum_priors > 0.0, "Train is broken: sum_priors=%f" % sum_priors

            for key, counts in self.probs[v]["counts"].iteritems():
                total = self.probs[v]["totals"][key]

                if self.log_mode:
                    self.probs[v]["probs"][key] = \
                        {val: math.log(counts[val] + priors[val]) - math.log(total + sum_priors)
                         for val in self.var_defs[v]["vals"]}
                else:
                    self.probs[v]["probs"][key] = \
                        {val: (counts[val] + priors[val]) / (total + sum_priors)
                         for val in self.var_defs[v]["vals"]}

        logger.info("Conditional probabilities calculated")

    #----------------------------------------------------------------------------------------------#
    # Serialization

    def dump(self):

        model = \
            {
                "var_names": self.var_names,
                "var_defs": self.var_defs,
                "log_mode": self.log_mode,
                "probs": self.probs
            }

        return json.dumps(model, indent=4, sort_keys=True)

    def save(self, filename):

        logger.info("Saving bayesnet to: %s", filename)

        if self.probs is None:
            raise BayesNetException("Model has not been trained")

        with open(filename, "w") as fout:
            model = \
                {
                    "var_names": self.var_names,
                    "var_defs": self.var_defs,
                    "log_mode": self.log_mode,
                    "probs": self.probs
                }

            json.dump(model, fout, indent=4, sort_keys=True)

    def load(self, filename):

        logger.info("Loading bayesnet from: %s", filename)

        with open(filename) as fin:
            model = json.load(fin)

            self.var_names = model["var_names"]
            self.var_defs = model["var_defs"]
            self.log_mode = model["log_mode"]
            self.probs = model["probs"]

            self.factors = None
            self.cached_factors = None

            for v in self.var_names:

                self.var_defs[v]["priors"] = \
                    {self._recast_val(val, self.var_defs[v]["type"]) : prior
                     for (val, prior) in self.var_defs[v]["priors"].iteritems()}

                for key in self.probs[v]["counts"].keys():

                    self.probs[v]["counts"][key] = \
                        {self._recast_val(val, self.var_defs[v]["type"]) : count
                         for (val, count) in self.probs[v]["counts"][key].iteritems()}
                    self.probs[v]["probs"][key] = \
                        {self._recast_val(val, self.var_defs[v]["type"]) : prob
                         for (val, prob) in self.probs[v]["probs"][key].iteritems()}

    #----------------------------------------------------------------------------------------------#
    # Inference

    def query(self, spec):

        logger.info("Querying bayesnet with: %s", spec)

        if self.probs is None:
            raise BayesNetException("Model has not been trained")

        self._init_factors()
        self._init_elimination_order()

        if "conditional_probability" in spec:
            query_data = spec["conditional_probability"]
            query_var_to_val = query_data["query_assignment"]
            evidence_var_to_val = query_data["evidence_assignment"]
            res = self._get_conditional_probability(query_var_to_val, evidence_var_to_val)

        elif "conditional_distribution" in spec:
            query_data = spec["conditional_distribution"]
            query_vars = query_data["query_variables"]
            evidence_var_to_val = query_data["evidence_assignment"]
            res = self._get_conditional_distribution(query_vars, evidence_var_to_val)

        elif "most_probable_explanation" in spec:
            query_data = spec["most_probable_explanation"]
            evidence_var_to_val = query_data["evidence_assignment"]
            res = self._get_most_probable_explanation(evidence_var_to_val)

        else:
            raise BayesNetException("Failed to parse query spec: %s" % spec)

        return res

    #----------------------------------------------------------------------------------------------#
    # Private - Training

    def _init_counts(self, var, dict_counts, dict_totals, cur_key, cur_parent):

        parents = self.var_defs[var]["parents"]

        if cur_parent == len(parents):
            dict_counts[cur_key] = {val: 0 for val in self.var_defs[var]["vals"]}
            dict_totals[cur_key] = 0
        else:
            for val in self.var_defs[parents[cur_parent]]["vals"]:
                new_key = "%s|%s" % (cur_key, str(val)) if cur_key else str(val)
                self._init_counts(var, dict_counts, dict_totals, new_key, cur_parent + 1)

    #----------------------------------------------------------------------------------------------#
    # Private - Serialization

    def _recast_val(self, val, var_type):

        if var_type == "str":
            val = str(val)
        elif var_type == "bool":
            val = bool(val)
        elif var_type == "int":
            val = int(val)
        elif var_type == "float":
            val = float(val)
        else:
            raise BayesNetException("Unrecognized variable type: %s" % var_type)

        return val

    #----------------------------------------------------------------------------------------------#
    # Private - Inference

    def _init_factors(self):

        if self.cached_factors:
            self._reset_factors()

        else:
            self.factors = []
            for v in self.var_names:
                self._add_factor(v)

            logger.info("Initialized factors")
            logger.debug("Factors:\n%s", self.factors)

            self.int_graph = self._get_interaction_graph()

            logger.info("Initialized interaction graph")
            logger.debug("Interactions:\n%s", self.int_graph)

            self._cache_factors()

    def _cache_factors(self):

        logger.info("Caching factors")

        self.cached_factors = [f.copy() for f in self.factors]
        self.cached_int_graph = {u: copy.copy(vs) for (u, vs) in self.int_graph.iteritems()}

    def _reset_factors(self):

        logger.info("Resetting to cached factors")

        self.factors = self.cached_factors
        self.int_graph = self.cached_int_graph

        self._cache_factors()

    def _get_interaction_graph(self):

        logger.info("Building interaction graph")

        res = {v: set() for v in self.var_names}
        for f in self.factors:
            for u, v in itertools.combinations(f.scope_vars, 2):
                for x, y in [(u, v), (v, u)]:
                    res[x].add(y)

        return res

    def _get_min_fill_order(self):

        logger.info("Building min-fill elimination ordering")

        n = len(self.var_names)
        if n <= 1:
            logger.info("Nothing to build for %d variables", n)
            return copy.copy(self.var_names)

        res = []

        for k in range(n):
            sel_v = None
            min_non_adjacent_ns = n * n

            non_adjacent_ns = {}
            for v in self.int_graph.keys():
                ns = [u for u in self.int_graph[v] if u in self.int_graph]
                non_adjacent_ns[v] = [(u, w)
                                      for (u, w) in itertools.combinations(ns, 2)
                                      if u not in self.int_graph[w]]
                if len(non_adjacent_ns[v]) < min_non_adjacent_ns:
                    sel_v = v
                    min_non_adjacent_ns = len(non_adjacent_ns[v])

            logger.info("\tVariable #%d selected for elimination: %s", k, sel_v)
            logger.info("\tNumber of non-adjacent neighbors: %d", min_non_adjacent_ns)
            res.append(sel_v)

            for (u, w) in non_adjacent_ns[sel_v]:
                for x, y in [(u, w), (w, u)]:
                    self.int_graph[x].add(y)

            del self.int_graph[sel_v]
            logger.debug("Remaining interactions:\n%s", self.int_graph)

        return res

    def _init_elimination_order(self, algo="min_fill"):

        if algo == "min_fill":
            var_order = self._get_min_fill_order()
        else:
            raise BayesNetException("Unrecognized variable elimination algo: %s", algo)

        self.elim_order = var_order
        logger.info("Initialized variable elimination order: %s", self.elim_order)

    def _add_factor(self, var):

        logger.info("Adding a new factor for: %s", var)

        vals_to_probs = {}
        scope_vars = [var] + self.var_defs[var]["parents"]

        parent_types = [self.var_defs[v]["type"] for v in self.var_defs[var]["parents"]]
        key_to_probs = self.probs[var]["probs"]

        for key, probs in key_to_probs.iteritems():
            if key:
                parent_vals = [self._recast_val(s, parent_types[i])
                               for i, s in enumerate(key.split("|"))]
            else:
                parent_vals = []

            for val, prob in probs.iteritems():
                if self.log_mode:
                    vals_to_probs[tuple([val] + parent_vals)] = math.exp(prob)
                else:
                    vals_to_probs[tuple([val] + parent_vals)] = prob

        factor = Factor(scope_vars, vals_to_probs)
        self.factors.append(factor)

    def _sum_product_eliminate_variable(self, var):

        logger.info("Running sum-product elimination of a single variable: %s", var)

        factors_with_var = [f for f in self.factors if var in f.scope_vars]
        factors_without_var = [f for f in self.factors if var not in f.scope_vars]

        if factors_with_var:
            f_new = reduce(lambda f, g: f.product(g, inplace=False),
                           factors_with_var[1:],
                           factors_with_var[0])
            f_new.sum_out(var)
            factors_without_var.append(f_new)

        self.factors = factors_without_var

    def _sum_product_eliminate_variables(self, vars):

        logger.info("Running sum-product variable elimination algorithm: V=%s", vars)

        vars_sorted = sorted(vars, key=lambda v: self.elim_order.index(v))
        logger.info("Elimination order: V=%s", vars_sorted)

        for var in vars_sorted:
            self._sum_product_eliminate_variable(var)

        f_new = reduce(lambda f, g: f.product(g, inplace=False),
                       self.factors[1:],
                       self.factors[0])
        self.factors = [f_new]

        return f_new

    def _max_product_eliminate_variable(self, var):

        logger.info("Running max-product elimination of a single variable: %s", var)

        factors_with_var = [f for f in self.factors if var in f.scope_vars]
        factors_without_var = [f for f in self.factors if var not in f.scope_vars]

        if factors_with_var:
            f_prod = reduce(lambda f, g: f.product(g, inplace=False),
                           factors_with_var[1:],
                           factors_with_var[0])
            f_new = f_prod.max_out(var, inplace=False)
            factors_without_var.append(f_new)

        else:
            raise BayesNetException("Attempt to eliminate unknown variable: %s" % var)

        self.factors = factors_without_var
        return f_prod

    def _max_product_eliminate_variables(self, vars):

        logger.info("Running max-product variable elimination algorithm: V=%s", vars)

        n_vars = len(vars)
        vars_sorted = sorted(vars, key=lambda v: self.elim_order.index(v))
        logger.info("Elimination order: V=%s", vars_sorted)

        f_prods = []
        for var in vars_sorted:
            f_prod = self._max_product_eliminate_variable(var)
            f_prods.append(f_prod)

        assert len(self.factors) > 0, \
               "MPE max-product elimination is broken: empty factor list"

        f_mpe = reduce(lambda f, g: f.product(g, inplace=False),
                       self.factors[1:],
                       self.factors[0])
        assert len(f_mpe.scope_vars) == 0,                                      \
               "MPE max-product elimination is broken: %s" % set(f_mpe.scope_vars)

        mpe_prob = f_mpe.vals_to_probs[()]
        assert 0.0 <= mpe_prob <= 1.0,                                          \
               "MPE max-product elimination is broken: p=%.4f" % mpe_prob

        logger.info("Calculating max-product traceback")

        mpe_var_to_val = {}
        for i in reversed(range(n_vars)):
            var_i = vars_sorted[i]
            f_prod = f_prods[i]

            elim_after_i = set(f_prod.scope_vars) - {var_i}
            max_assignment = {var: mpe_var_to_val[var] for var in elim_after_i}
            max_i = f_prod.arg_max(var_i, max_assignment)
            mpe_var_to_val[var_i] = max_i

        logger.info("Max-product traceback complete, best assignment: %s", mpe_var_to_val)
        logger.info("Probability of best assignment: %.4f", mpe_prob)

        return mpe_var_to_val, mpe_prob

    def _get_conditional_distribution(self, query_vars, evidence_var_to_val):

        logger.info("Building conditional probability distribution: Q=%s, E=%s",
                    query_vars, evidence_var_to_val)

        vars_out = [v for v in self.var_names
                    if v not in query_vars and v not in evidence_var_to_val]

        for (var, val) in evidence_var_to_val.iteritems():
            for f in self.factors:
                if var in f.scope_vars:
                    f.reduce(var, val)

        self.factors = [f for f in self.factors if f.scope_vars]

        logger.debug("Pre-processed factors before elimination:\n%s", self.factors)

        f_res = self._sum_product_eliminate_variables(vars_out)
        f_res.normalize()

        return f_res

    def _get_conditional_probability(self, query_var_to_val, evidence_var_to_val):

        logger.info("Processing conditional probability query: Q=%s, E=%s",
                    query_var_to_val, evidence_var_to_val)

        query_vars = query_var_to_val.keys()

        f = self._get_conditional_distribution(query_vars, evidence_var_to_val)
        assert set(f.scope_vars) == set(query_vars), \
               "Conditional distribution is broken: %s != %s" % (f.scope_vars, query_vars)

        query_vals = tuple([query_var_to_val[v] for v in f.scope_vars])
        assert query_vals in f.vals_to_probs, \
               "Conditional distribution does not contain values: %s" % query_vals

        p = f.vals_to_probs[query_vals]

        logger.info("Calculated conditional probability: %f", p)

        return p

    def _get_most_probable_explanation(self, evidence_var_to_val):

        logger.info("Processing most probable explanation query: E=%s",
                    evidence_var_to_val)

        vars_out = [v for v in self.var_names
                    if v not in evidence_var_to_val]

        for (var, val) in evidence_var_to_val.iteritems():
            for f in self.factors:
                if var in f.scope_vars:
                    f.reduce(var, val)

        logger.debug("Pre-processed factors before elimination:\n%s", self.factors)

        mpe_var_to_val, mpe_prob = self._max_product_eliminate_variables(vars_out)
        return mpe_var_to_val, mpe_prob

#==================================================================================================#

if __name__ == "__main__":
    pass
